{
   k10ctl -- AMD Family 10h (aka K10) P-State, frequency and voltage modification utility
   Copyright (C) 2009 Stefan Ziegenbalg
   http://www.ztex.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
}

{$mode objfpc}
uses
{$ifdef VER2}
  oldlinux,
{$else}
  linux,
{$endif}  
  errors;


{ $define test}

const PSTATE_CTL    = $C0010062;
      PSTATE_STATUS = $C0010063;
      PSTATE_CFG    = $C0010064;
      PSTATE_NUM    = 5;
      COFVID_CTL    = $C0010070;
      COFVID_STATUS = $C0010071;
      
      pvi  : boolean = false;
      
type TMSRBuf = packed record
      lo,hi : dword;
      end;
     
     TUF = record
      nbVid, nbDid            : longint;
      cpuVid, cpuDid, cpuFid  : longint;
      pstate                  : longint;
      cpuMult,UNb,UCpu        : double;
      end;



procedure error(msg:ansistring);
begin
writeln(stderr,msg);
halt(2);
end;


function int2str(i:longint):string[15];
begin
str(i,int2str);
end;


procedure setBits(var d:dword; bit0,bit1:longint; v:longint);
var i  : longint;
begin
if bit0>bit1 then
  begin
  i:=bit0;
  bit0:=bit1;
  bit1:=i;
  end;
{writeln(hexstr( (1 shl bit0) - (2 shl bit1) - 1 ,8));
writeln(hexstr(v shl bit0, 8));
writeln(2 shl (bit1-bit0)); }
if (v>=0) and (v < 2 shl (bit1-bit0)) then d:=(d and ( (1 shl bit0) - (2 shl bit1) - 1) ) or (v shl bit0);
end;


const llseek_pos : int64 = 0;

function llseek(fd:longint; off:dword):boolean;
var regs   : SysCallregs;
begin
regs.reg2:=fd;
regs.reg3:=0;
regs.reg4:=off;
regs.reg5:=longint(@llseek_pos);
regs.reg6:=seek_set;
result:=SysCall(syscall_nr__llseek,regs) = 0;
linuxerror:=errno;
end;


procedure rdmsr(fd:longint; msr:dword; var buf:TMsrBuf);
begin
if not llseek(fd, msr) then error('Error seeking to MSR '+hexstr(msr,8)+': '+StrError(linuxerror) );
if fdread(fd,buf,8)<>8 then error('Error reading MSR '+hexstr(msr,8)+': '+StrError(linuxerror) );
end;


procedure wrmsr(fd:longint; msr:dword; var buf:TMsrBuf);
begin
if not llseek(fd, msr) then error('Error seeking to MSR '+hexstr(msr,8)+': '+StrError(linuxerror) );
if fdwrite(fd,buf,8)<>8 then error('Error writeing to MSR '+hexstr(msr,8)+': '+StrError(linuxerror) );
end;


procedure paramerr(msg: ansistring);
begin
if msg<>'' then 
  begin
  writeln(stderr,msg);
  writeln(stderr);
  end;
writeln(stderr,'Warning:');
writeln(stderr,'  USE THIS PROGRAM AT YOU OWN RISK. IT MAY DAMAGE YOUR HARDWARE.');
writeln(stderr);
writeln(stderr,'Usage: ');
writeln(stderr,'  k10ctl -h                                     This help');
writeln(stderr,'  k10ctl <cpu>[-<cpun>]                         Print a lot of information about CPU(s) <cpu>(-<cpun>)');
writeln(stderr,'  k10ctl <cpu>[-<cpun>] -a                      Print DSDT template for CPU(s) <cpu>(-<cpun>) using P-State MSR''s');
writeln(stderr,'  k10ctl <cpu>[-<cpun>] <p-state>               Set the P-State for CPU(s) <cpu>(-<cpun>)');
writeln(stderr,'  k10ctl <cpu>[-<cpun>] <p-state> [parameters]  Configure the P-State for CPU(s) <cpu>(-<cpun>), see below');
writeln(stderr);
writeln(stderr,'Parameters for the P-State configuration:');
writeln(stderr,'  -nv <NbVid>    Nothbridge VID (0..127)');
writeln(stderr,'  -nd <NbDid>    Nothbridge DID (0..1)');
writeln(stderr,'  -cv <CpuVid>   CPU VID (0..127)');
writeln(stderr,'  -cd <CpuDid>   CPU DID (0..7)');
writeln(stderr,'  -cf <CpuFid>   CPU FID (0..63)');
writeln(stderr,'  -p <mW>        Power dissipation in mW');
writeln(stderr);
writeln(stderr,'Examples: ');
writeln(stderr,'  k10ctl 0-3                                    Print a lot of information about cores 0-3');
writeln(stderr,'  k10ctl 0-3 0 -nv 46 -cv 22 -cd 0 -cf 16	Set NbVid=46, CpuVid=22, CpuDiD=0, CpuFid=16 for CPU''s 0-3 and P-State 0');
writeln(stderr,'  k10ctl 0-3 0     				Set P-State 0 for CPU''s 0-3');
writeln(stderr,'  k10ctl 0 -a 		                     	Print a DSDT template for CPU 0');
if msg<>'' then halt(1);
halt(0);
end;


function val2(const s:shortstring; var d:longint):boolean;
var i,r  : longint;
begin
val(s,r,i);
result:=i=0;
if result then d:=r;
end;


procedure val3(const s:shortstring; var d:longint);
var i : longint;
begin
val(s,d,i);
if i<>0 then paramerr('Number expected: `'+s+'''');
end;

procedure val3(const s:shortstring; var d:longint; max:longint);
var i  : longint;
begin
val(s,d,i);
if (i<>0) or (d<0) or (d>max) then paramerr('Number between 0 and '+int2str(max)+' expected: `'+s+'''');
end;


function vid2U(i:longint):double;
begin
if pvi 
  then begin
       if i>63 then i:=63;
       if i<32 then result:=1550 - i*25
               else result:=1162.5 - i*12.5;
       end
  else begin
       if i>124 then i:=124;
       result:=1550 - i*12.5;
       end;
       
end;


function getI(d:dword):double;
const iFact : array[0..3] of double = (1, 0.1, 0.01, 0.001);
begin
result:=(d and 255) * iFact[(d shr 8) and 3];
end;


function getUF(d:dword):TUF;
begin
result.nbVid:=d shr 25;
result.nbDid:=(d shr 22) and 1;
result.pstate:=(d shr 16) and 7;
result.cpuVid:=(d shr 9) and 127;
result.cpuDid:=(d shr 6) and 7;
result.cpuFid:=d and 63;
result.cpuMult:=(result.cpuFid+16)/(2 shl result.cpuDid);
result.UNb:=vid2U(result.nbVid);
result.UCpu:=vid2U(result.cpuVid);
end;


function readPState(fd, pstate:dword; var buf:TMSRBuf):boolean;
begin
if (pstate<0) or (pstate>4) then error('Error: P-State out of range [0..4]: '+int2str(pstate));
rdmsr(fd, PSTATE_CFG+pstate, buf);
result:=buf.hi and $80000000<>0;
end; 


procedure info(fd:longint);
var msrfn     : ansistring;
    i         : longint;
    buf0,buf1 : TMSRBuf;
    uf        : TUF;
begin
rdmsr(fd, PSTATE_STATUS, buf0);
rdmsr(fd, COFVID_STATUS, buf1);

writeln('Current P-State: ',buf0.lo and 7,'      Fastest P-State: ', (buf1.hi shr 24) and 7);

writeln('            ', 'NbVid':8, 'NbDid':8, 'CpuVid':8, 'CpuDid':8, 'CpuFid':8, 	'UNb':14, 'CpuMult':10, 'UCpu':10, 'PCore':10);
for i:=0 to PSTATE_NUM-1 do
  begin
  if readPState(fd, i, buf0) then
    begin
    uf:=getUF(buf0.lo); 
    writeln('P-State ',i,':  ', uf.nbVid:8, uf.nbDid:8, uf.cpuVid:8, uf.cpuDid:8, uf.cpuFid:8,	uf.UNb:12:1,'mV', uf.cpuMult:10:5, uf.UCpu:8:1,'mV', round(getI(buf0.hi) *uf.UCpu):8,'mW');
    end;
  end; 

i:=(buf1.hi shr 10) and 127;
if (i=0) or (i>124) then i:=124;
writeln( 'Low Limit:  ', i:8, '1':8, i:8, vid2U(i):28:1,'mV', '0.50000':10, vid2U(i):8:1,'mV' );
i:=(buf1.hi shr 3) and 127;
writeln( 'High Limit: ', i:8, '1':8, i:8, vid2U(i):28:1,'mV', 0.5*((buf1.hi shr 17) and 63):10:5, vid2U(i):8:1,'mV' );

rdmsr(fd, COFVID_CTL, buf0);
uf:=getUF(buf0.lo); 
writeln('Target:     ', uf.nbVid:8, uf.nbDid:8, uf.cpuVid:8, uf.cpuDid:8, uf.cpuFid:8,	uf.UNb:12:1,'mV', uf.cpuMult:10:5, uf.UCpu:8:1,'mV');
uf:=getUF(buf1.lo); 
writeln('Current:    ', uf.nbVid:8, uf.nbDid:8, uf.cpuVid:8, uf.cpuDid:8, uf.cpuFid:8,	uf.UNb:12:1,'mV', uf.cpuMult:10:5, uf.UCpu:8:1,'mV');
end;


procedure DSDTInfo(fd:longint);
var buf  : TMSRBuf;
    i,j  : longint;
    uf   : TUF;
begin
j:=0;
for i:=0 to PSTATE_NUM-1 do
  if readPState(fd, i, buf) then j+=1;

writeln('	{');
writeln('            Name (_PPC, 0x00)');
writeln;
writeln('            Name (_PCT, Package (0x02)');
writeln('            {');
writeln('                ResourceTemplate ()');
writeln('                {');
writeln('                    Register (FFixedHW, 	// COFVID_CTL');
writeln('                        0x20,              	// Bit Width');
writeln('                        0x00,               	// Bit Offset');
writeln('                        0x',hexstr(COFVID_CTL,16),')	// Address');
writeln('                },');
writeln;
writeln('                ResourceTemplate ()');
writeln('                {');
writeln('                    Register (FFixedHW, 	// COFVID_STATUS');
writeln('                        0x20,	               	// Bit Width');
writeln('                        0x00,    		// Bit Offset');
writeln('                        0x',hexstr(COFVID_STATUS,16),') 	// Address');
writeln('                }');
writeln('            })');
writeln;
writeln('            Name (_PSS, Package (',j,')');
writeln('            {');
for i:=0 to PSTATE_NUM-1 do
  if readPState(fd, i, buf) then
    begin
    uf:=getUF(buf.lo); 
    setBits(buf.lo, 18,16, i);
    writeln('                Package (0x06)		// P-State ',i);
    writeln('                {');
    writeln('                    ',round(uf.cpuMult*200),', 		// f in MHz');
    writeln('                    ',round(getI(buf.hi)*uf.UCpu),', 		// P in mW');
    writeln('                    10,			// Transition latency in us');
    writeln('                    10,			// Bus Master latency in us');
    writeln('                    0x0',hexstr(buf.lo,8),',	// value written to COFVID_CTL; nbVid=', 
	    uf.nbVid ,' (',uf.UNb:0:1,'mV),  nbDid=',uf.nbDid,',  cpuVid=',uf.cpuVid, ' (',uf.UCpu:0:1,'mV),  cpuDid=',
	    uf.cpuDid,',  cpuFid=',uf.cpuFid,' (cpuMult=',uf.cpuMult:0:5,')');
    writeln('                    0x0',hexstr(buf.lo,8),',	// value of COFVID_STATUS after successful transition');
    writeln('                },');
    end;
writeln('            })');
writeln('	}');
writeln(stderr,'Please edit frequency and power dissipation by hand if necessary.');
end;


procedure cfgPState(fd:longint; pstate:dword; mw,nv,nd,cv,cd,cf : longint);
var buf : TMSRBuf;
    i   : longint;
    r   : double;
{$ifdef test}
    uf  : TUF;
{$endif}
begin
if not readPState(fd, pstate, buf) then 
  begin
  if pstate=0 then error('Error: Invalid P-State 0');
  if not readPState(fd, pstate-1, buf) then
    begin
    i:=0;
    while (i<PSTATE_NUM-1) and readPState(fd, i, buf) do
      i+=1;
    error('Invalid P-State. Use P-States < '+int2str(i)+' to overwrite an existing one or use P-State '+int2str(i)+' to create a new one.');
    end; 
  end;

buf.hi:=buf.hi or $80000000;
setBits(buf.lo, 31,25, nv);
setBits(buf.lo, 22,22, nd);
setBits(buf.lo, 15,9, cv);
setBits(buf.lo, 8,6, cd);
setBits(buf.lo, 5,0, cf);

if mw>=0 then
  begin
  r:=mw / (vid2U((buf.lo shr 9) and 127)+1e-10);
  i:=0;
  while (r<25.55) and (i<2) do
    begin
    r*=10;
    i+=1;
    end;
  if r>255 then r:=255;
  setBits(buf.hi, 9,8, i);
  setBits(buf.hi, 7,0, round(r));
  end;
  
{$ifdef test}
uf:=getUF(buf.lo); 
writeln('            ', 'NbVid':8, 'NbDid':8, 'CpuVid':8, 'CpuDid':8, 'CpuFid':8, 	'UNb':14, 'cpuMult':10, 'UCpu':10, 'PCore':10);
writeln('P-State ',pstate,':  ', uf.nbVid:8, uf.nbDid:8, uf.cpuVid:8, uf.cpuDid:8, uf.cpuFid:8,	uf.UNb:12:1,'mV', uf.cpuMult:10:5, uf.UCpu:8:1,'mV', round(getI(buf.hi) *uf.UCpu):8,'mW');
{$endif}

wrmsr(fd, PSTATE_CFG+pstate, buf);
end; 


procedure setPState(fd:longint; pstate:longint);
var buf0,buf1  : TMSRBuf;
begin
if not readPState(fd, pstate, buf0) then error('Invalid P-State: '+int2str(pstate));
rdmsr(fd, COFVID_CTL, buf1);
buf1.lo:=buf0.lo;
setBits(buf1.lo, 18,16, pstate);
wrmsr(fd, COFVID_CTL, buf1);
end;

var fd,i              : longint;
    s,t               : shortstring;
    msrfn             : ansistring;
    cpua,cpuz         : longint;
    pciconf           : array[0..255] of byte;
    
    
    pstate            : longint;
    mw,nv,nd,cv,cd,cf : longint;

begin
for i:=1 to paramcount do
  if paramstr(i)='-h' then paramerr('');
if paramcount<1 then paramerr('');
  
s:=paramstr(1);
i:=2;
while (i<length(s)) and (s[i]<>'-') do
  i+=1;
if i<length(s) then 
    begin
    val3(copy(s,1,i-1),cpua);
    val3(copy(s,i+1,length(s)-i),cpuz);
    if cpuz<cpua then
      begin
      i:=cpuz;
      cpuz:=cpua;
      cpua:=i;
      end;
    end  
  else 
    begin
    val3(s,cpua);
    cpuz:=cpua;
    end;

fd:=fdopen('/proc/bus/pci/00/18.3',open_rdonly);
if (fd<0) or (linuxerror<>0) then error('Error opening /proc/bus/pci/00/18.3: '+StrError(linuxerror));
if fdread(fd,pciconf,256)<>256 then error('Error reading PCI configuration space: '+StrError(linuxerror) );
fdclose(fd);
// check vendor and device ID
//writeln(hexstr(pciconf[3],2), hexstr(pciconf[2],2),' ' ,hexstr(pciconf[1],2), hexstr(pciconf[0],2));
if (pciconf[3]<>$12) or (pciconf[2]<>$3) or (pciconf[1]<>$10) or (pciconf[0]<>$22) then error('Invalid PCI configuration space');
// read PVI or SVI mode
pvi:=pciconf[$a1] and 1 = 1;;

if paramcount=1 then  
    begin
    for i:=cpua to cpuz do
      begin
      msrfn:='/dev/cpu/'+int2str(i)+'/msr';
      fd:=fdopen(msrfn,open_rdonly);
      if (fd<0) or (linuxerror<>0) then error('Error opening `'+msrfn+''': '+StrError(linuxerror));
      write('VID interface mode: ');
      if pvi 
          then writeln('parallel')
          else writeln('serial');
      writeln;
      
      writeln('CPU',i);
      info(fd);
      writeln;
      fdclose(fd);
      end;
    end
 else if paramcount=2 then
    begin
    if paramstr(2)='-a' then
        begin
        for i:=cpua to cpuz do
   	  begin
          msrfn:='/dev/cpu/'+int2str(i)+'/msr';
          fd:=fdopen(msrfn,open_rdonly);
          if (fd<0) or (linuxerror<>0) then error('Error opening `'+msrfn+''': '+StrError(linuxerror));
          DSDTInfo(fd);
          fdclose(fd);
	  end;
        end
      else 
        begin
	val3(paramstr(2),pstate,4);
	for i:=cpua to cpuz do
          begin
          msrfn:='/dev/cpu/'+int2str(i)+'/msr';
          fd:=fdopen(msrfn,open_rdwr);
          if (fd<0) or (linuxerror<>0) then error('Error opening `'+msrfn+''': '+StrError(linuxerror));
	  setPState(fd,pstate);
          fdclose(fd);
          end;
	end;
    end
  else 
    begin
    val3(paramstr(2),pstate,4);
    
    mw:=-1;
    nv:=-1;
    nd:=-1;
    cv:=-1;
    cd:=-1;
    cf:=-1;
    
    i:=3;
    while i<=paramcount do
      begin
      s:=paramstr(i);
      if i<paramcount 
        then t:=paramstr(i+1)
	else paramerr('Value expected after '+s);
      if s='-p' then val3(t,mw)
        else if s='-nv' then val3(t,nv,124)
        else if s='-nd' then val3(t,nd,1)
        else if s='-cv' then val3(t,cv,124)
        else if s='-cd' then val3(t,cd,7)
        else if s='-cf' then val3(t,cf,63)
	else paramerr('Invalid parameter: `'+s+'''');
      i+=2;
      end;
      
    for i:=cpua to cpuz do
      begin
      msrfn:='/dev/cpu/'+int2str(i)+'/msr';
      fd:=fdopen(msrfn,open_rdwr);
      if (fd<0) or (linuxerror<>0) then error('Error opening `'+msrfn+''': '+StrError(linuxerror));
      cfgPState(fd, pstate, mw,nv,nd,cv,cd,cf);
      fdclose(fd);
      end;
    end;
end.

